from birder.net.ssl.barlow_twins import BarlowTwins
from birder.net.ssl.byol import BYOL
from birder.net.ssl.capi import CAPIStudent
from birder.net.ssl.capi import CAPITeacher
from birder.net.ssl.data2vec import Data2Vec
from birder.net.ssl.data2vec2 import Data2Vec2
from birder.net.ssl.dino_v1 import DINO_v1
from birder.net.ssl.dino_v2 import DINOv2Student
from birder.net.ssl.dino_v2 import DINOv2Teacher
from birder.net.ssl.i_jepa import I_JEPA
from birder.net.ssl.ibot import iBOT
from birder.net.ssl.mmcr import MMCR
from birder.net.ssl.simclr import SimCLR
from birder.net.ssl.sscd import SSCD
from birder.net.ssl.vicreg import VICReg

__all__ = [
    "BarlowTwins",
    "BYOL",
    "CAPIStudent",
    "CAPITeacher",
    "Data2Vec",
    "Data2Vec2",
    "DINO_v1",
    "DINOv2Student",
    "DINOv2Teacher",
    "I_JEPA",
    "iBOT",
    "MMCR",
    "SimCLR",
    "SSCD",
    "VICReg",
]
